"""
pyrad.graph.plots_spectra
=========================

Functions to plot spectral data

.. autosummary::
    :toctree: generated/

    plot_range_Doppler
    plot_angle_Doppler
    plot_time_Doppler
    plot_Doppler
    plot_complex_range_Doppler
    plot_complex_angle_Doppler
    plot_complex_time_Doppler
    plot_amp_phase_range_Doppler
    plot_amp_phase_angle_Doppler
    plot_amp_phase_time_Doppler
    plot_complex_Doppler
    plot_amp_phase_Doppler
    _create_irregular_grid
    _adapt_data_to_irregular_grid

"""

from .plots_aux import generate_angle_Doppler_title
from .plots_aux import generate_complex_Doppler_title
from .plots_aux import generate_complex_range_Doppler_title
from .plots_aux import get_colobar_label, get_norm
import pyart
import matplotlib.pyplot as plt
import numpy as np

import matplotlib as mpl

mpl.use("Agg")

# Increase a bit font size
mpl.rcParams.update({"font.size": 16})
mpl.rcParams.update({"font.family": "sans-serif"})


def plot_range_Doppler(
    spectra,
    field_name,
    ray,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
):
    """
    Makes a range-Doppler plot

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray : int
        ray index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    titl : str or None
        The plot title
    clabel : str or None
        The color bar label
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    ymin = prdcfg["spectraImageConfig"]["ymin"]
    ymax = prdcfg["spectraImageConfig"]["ymax"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"

    xres = np.abs(xaxis[1] - xaxis[0])

    yaxis = spectra.range["data"] / 1000.0
    yres = np.abs(yaxis[1] - yaxis[0])
    ylabel = "range (km)"

    xaxis_lim = np.append(xaxis - xres / 2, xaxis[-1] + xres / 2)
    yaxis_lim = np.append(yaxis - yres / 2, yaxis[-1] + yres / 2)

    field_2D = spectra.fields[field_name]["data"][ray, :, 0 : xaxis.size]

    X, Y = np.meshgrid(xaxis_lim, yaxis_lim)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min(field_2D)
        if vmax is None:
            vmax = np.ma.max(field_2D)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_range_Doppler_title(
            spectra, field_name, ray, datetime_format=None
        )

    ax = fig.add_subplot(111)
    cax = ax.pcolormesh(X, Y, field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(titl)

    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])
    if not (ymin is None or ymax is None):
        ax.set_ylim([ymin, ymax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_angle_Doppler(
    spectra,
    field_name,
    ang,
    ind_rays,
    ind_rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="centre",
    along_azi=True,
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
):
    """
    Makes an angle-Doppler plot

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ang : float
        The fixed angle
    ind_rays : 1D int array
        The indices of the rays to plot
    ind_rng : int
        The index of the range to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    along_azi : bool
        If true the plot is performed along azimuth. If false it is performed
        along elevation
    titl : str or None
        The plot title
    clabel : str or None
        The color bar label
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ind_rays, :]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ind_rays, :]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((ind_rays.size, spectra.npulses_max))
        for ray, ind_ray in enumerate(ind_rays):
            npuls = spectra.npulses["data"][ind_ray]
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "Angle (deg)"

    if along_azi:
        yaxis = spectra.azimuth["data"][ind_rays]
    else:
        yaxis = spectra.elevation["data"][ind_rays]

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, yaxis, yaxis_pos=yaxis_pos
    )

    data = spectra.fields[field_name]["data"][ind_rays, :, :]
    data = data[:, ind_rng, :]
    zpoints = _adapt_data_to_irregular_grid(
        data, xaxis_size, xaxis_lim.shape[0], xaxis_lim.shape[1]
    )

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min(zpoints)
        if vmax is None:
            vmax = np.ma.max(zpoints)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_angle_Doppler_title(
            spectra, field_name, ang, ind_rng, along_azi=along_azi, datetime_format=None
        )

    ax = fig.add_subplot(111)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, zpoints, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(titl)
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_time_Doppler(
    spectra,
    field_name,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="start",
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
):
    """
    Makes a time-Doppler plot

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    titl : str or None
        The plot title
    clabel : str or None
        The color bar label
    vmin, vmax : float or None
        The value limits
    xmin, xmax, ymin, ymax : float or None
        The axis limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((spectra.nrays, spectra.npulses_max))
        for ray, npuls in enumerate(spectra.npulses["data"]):
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "time (s from start time)"

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, spectra.time["data"], yaxis_pos=yaxis_pos
    )

    zpoints = _adapt_data_to_irregular_grid(
        np.ma.squeeze(spectra.fields[field_name]["data"], axis=1),
        xaxis_size,
        xaxis_lim.shape[0],
        xaxis_lim.shape[1],
    )

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min(zpoints)
        if vmax is None:
            vmax = np.ma.max(zpoints)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, 0, 0, datetime_format=None
        )

    ax = fig.add_subplot(111)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, zpoints, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(titl)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_Doppler(
    spectra,
    field_name,
    ray,
    rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    ylabel=None,
    titl=None,
    vmin=None,
    vmax=None,
):
    """
    Makes a Doppler plot

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray, rng : int
        ray and rng index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    ylabel : str or None
        The label of the y-axis
    titl : str or None
        The plot title
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"

    field = spectra.fields[field_name]["data"][ray, rng, 0 : xaxis.size]

    # display data
    if field_name is not None:
        if ylabel is None:
            field_dict = pyart.config.get_metadata(field_name)
            ylabel = get_colobar_label(field_dict, field_name)
        if vmin is None or vmax is None:
            vmin, vmax = pyart.config.get_field_limits(field_name)
    else:
        if vmin is None:
            vmin = np.ma.min(field)
        if vmax is None:
            vmax = np.ma.max(field)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, ray, rng, datetime_format=None
        )

    ax = fig.add_subplot(111)
    ax.plot(xaxis, field, marker="x")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_ylim(bottom=vmin, top=vmax)
    ax.set_xlim([xaxis[0], xaxis[-1]])
    ax.set_title(titl)

    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    # Turn on the grid
    ax.grid()

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_complex_range_Doppler(
    spectra,
    field_name,
    ray,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
):
    """
    Makes a complex range-Doppler plot. Plotting separately the real and the
    imaginary part

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray : int
        ray index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    titl : str or None
        The plot title
    clabel : str or None
        The label of color bar
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    ymin = prdcfg["spectraImageConfig"]["ymin"]
    ymax = prdcfg["spectraImageConfig"]["ymax"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"

    xres = np.abs(xaxis[1] - xaxis[0])

    yaxis = spectra.range["data"] / 1000.0
    yres = np.abs(yaxis[1] - yaxis[0])
    ylabel = "range (km)"

    xaxis_lim = np.append(xaxis - xres / 2, xaxis[-1] + xres / 2)
    yaxis_lim = np.append(yaxis - yres / 2, yaxis[-1] + yres / 2)

    field_2D = spectra.fields[field_name]["data"][ray, :, 0 : xaxis.size]

    re_field_2D = np.real(field_2D)
    im_field_2D = np.imag(field_2D)

    X, Y = np.meshgrid(xaxis_lim, yaxis_lim)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min([np.ma.min(re_field_2D), np.ma.min(im_field_2D)])
        if vmax is None:
            vmax = np.ma.max([np.ma.max(re_field_2D), np.ma.max(im_field_2D)])

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_range_Doppler_title(
            spectra, field_name, ray, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolormesh(X, Y, re_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Real part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])
    if not (ymin is None or ymax is None):
        ax.set_ylim([ymin, ymax])

    ax = fig.add_subplot(122)
    cax = ax.pcolormesh(X, Y, im_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Imaginary part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])
    if not (ymin is None or ymax is None):
        ax.set_ylim([ymin, ymax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_complex_angle_Doppler(
    spectra,
    field_name,
    ang,
    ind_rays,
    ind_rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="centre",
    along_azi=True,
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
):
    """
    Makes an angle-Doppler plot of complex spectra

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ang : float
        The fixed angle
    ind_rays : 1D int array
        The indices of the rays to plot
    ind_rng : int
        The index of the range to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    along_azi : bool
        If true the plot is performed along azimuth. If false it is performed
        along elevation
    titl : str or None
        The plot title
    clabel : str or None
        The color bar label
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ind_rays, :]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ind_rays, :]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((ind_rays.size, spectra.npulses_max))
        for ray, ind_ray in enumerate(ind_rays):
            npuls = spectra.npulses["data"][ind_ray]
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "Angle (deg)"

    if along_azi:
        yaxis = spectra.azimuth["data"][ind_rays]
    else:
        yaxis = spectra.elevation["data"][ind_rays]

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, yaxis, yaxis_pos=yaxis_pos
    )

    data = spectra.fields[field_name]["data"][ind_rays, :, :]
    data = data[:, ind_rng, :]
    field_2D = _adapt_data_to_irregular_grid(
        data, xaxis_size, xaxis_lim.shape[0], xaxis_lim.shape[1]
    )

    re_field_2D = np.real(field_2D)
    im_field_2D = np.imag(field_2D)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min([np.ma.min(re_field_2D), np.ma.min(im_field_2D)])
        if vmax is None:
            vmax = np.ma.max([np.ma.max(re_field_2D), np.ma.max(im_field_2D)])

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_angle_Doppler_title(
            spectra, field_name, ang, ind_rng, along_azi=along_azi, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, re_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Real part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    ax = fig.add_subplot(122)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, im_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Imaginary part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_complex_time_Doppler(
    spectra,
    field_name,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="start",
    titl=None,
    clabel=None,
    vmin=None,
    vmax=None,
):
    """
    Makes a complex time-Doppler plot. Plotting separately the real and the
    imaginary part

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    titl : str or None
        The plot title
    clabel : str or None
        The label of color bar
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((spectra.nrays, spectra.npulses_max))
        for ray, npuls in enumerate(spectra.npulses["data"]):
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "time (s from start time)"

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, spectra.time["data"], yaxis_pos=yaxis_pos
    )

    field_2D = _adapt_data_to_irregular_grid(
        np.ma.squeeze(spectra.fields[field_name]["data"], axis=1),
        xaxis_size,
        xaxis_lim.shape[0],
        xaxis_lim.shape[1],
    )

    re_field_2D = np.real(field_2D)
    im_field_2D = np.imag(field_2D)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        field_dict = pyart.config.get_metadata(field_name)
        if clabel is None:
            clabel = get_colobar_label(field_dict, field_name)

        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if vmin is None or vmax is None:
            vmin = vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                vmin, vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None
    else:
        if clabel is None:
            clabel = "value"
        if vmin is None:
            vmin = np.ma.min([np.ma.min(re_field_2D), np.ma.min(im_field_2D)])
        if vmax is None:
            vmax = np.ma.max([np.ma.max(re_field_2D), np.ma.max(im_field_2D)])

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, 0, 0, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, re_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Real part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    ax = fig.add_subplot(122)
    cax = ax.pcolor(
        xaxis_lim, yaxis_lim, im_field_2D, cmap=cmap, vmin=vmin, vmax=vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Imaginary part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label(clabel)

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_amp_phase_range_Doppler(
    spectra,
    field_name,
    ray,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    titl=None,
    ampli_vmin=None,
    ampli_vmax=None,
    phase_vmin=None,
    phase_vmax=None,
):
    """
    Makes a complex range-Doppler plot plotting separately the module and the
    phase of the signal

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray : int
        ray index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    titl : str or None
        The plot title
    ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    ymin = prdcfg["spectraImageConfig"]["ymin"]
    ymax = prdcfg["spectraImageConfig"]["ymax"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"
    xres = np.abs(xaxis[1] - xaxis[0])

    yaxis = spectra.range["data"] / 1000.0
    yres = np.abs(yaxis[1] - yaxis[0])
    ylabel = "range (km)"

    xaxis_lim = np.append(xaxis - xres / 2, xaxis[-1] + xres / 2)
    yaxis_lim = np.append(yaxis - yres / 2, yaxis[-1] + yres / 2)

    field_2D = spectra.fields[field_name]["data"][ray, :, 0 : xaxis.size]

    ampli_field_2D = np.ma.abs(field_2D)
    phase_field_2D = np.ma.angle(field_2D, deg=True)

    X, Y = np.meshgrid(xaxis_lim, yaxis_lim)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if ampli_vmin is None or ampli_vmax is None:
            ampli_vmin = ampli_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                ampli_vmin, ampli_vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None

        if phase_vmin is None or phase_vmax is None:
            phase_vmin = phase_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                phase_vmin, phase_vmax = (-180.0, 180.0)
        else:
            norm = None
    else:
        if ampli_vmin is None:
            ampli_vmin = np.ma.min(ampli_field_2D)
        if ampli_vmax is None:
            ampli_vmax = np.ma.max(ampli_field_2D)
        if phase_vmin is None:
            phase_vmin = np.ma.min(phase_field_2D)
        if phase_vmax is None:
            phase_vmax = np.ma.max(phase_field_2D)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_range_Doppler_title(
            spectra, field_name, ray, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolormesh(
        X, Y, ampli_field_2D, cmap=cmap, vmin=ampli_vmin, vmax=ampli_vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Amplitude")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])
    if not (ymin is None or ymax is None):
        ax.set_ylim([ymin, ymax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Amplitude (-)")

    ax = fig.add_subplot(122)
    cax = ax.pcolormesh(
        X, Y, phase_field_2D, cmap=cmap, vmin=phase_vmin, vmax=phase_vmax, norm=norm
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Phase")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])
    if not (ymin is None or ymax is None):
        ax.set_ylim([ymin, ymax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Phase (deg)")

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_amp_phase_angle_Doppler(
    spectra,
    field_name,
    ang,
    ind_rays,
    ind_rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="centre",
    along_azi=True,
    titl=None,
    ampli_vmin=None,
    ampli_vmax=None,
    phase_vmin=None,
    phase_vmax=None,
):
    """
    Makes an angle-Doppler plot of complex spectra

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ang : float
        The fixed angle
    ind_rays : 1D int array
        The indices of the rays to plot
    ind_rng : int
        The index of the range to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    along_azi : bool
        If true the plot is performed along azimuth. If false it is performed
        along elevation
    titl : str or None
        The plot title
    ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ind_rays, :]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ind_rays, :]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((ind_rays.size, spectra.npulses_max))
        for ray, ind_ray in enumerate(ind_rays):
            npuls = spectra.npulses["data"][ind_ray]
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "Angle (deg)"

    if along_azi:
        yaxis = spectra.azimuth["data"][ind_rays]
    else:
        yaxis = spectra.elevation["data"][ind_rays]

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, yaxis, yaxis_pos=yaxis_pos
    )

    data = spectra.fields[field_name]["data"][ind_rays, :, :]
    data = data[:, ind_rng, :]
    field_2D = _adapt_data_to_irregular_grid(
        data, xaxis_size, xaxis_lim.shape[0], xaxis_lim.shape[1]
    )

    ampli_field_2D = np.ma.abs(field_2D)
    phase_field_2D = np.ma.angle(field_2D, deg=True)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if ampli_vmin is None or ampli_vmax is None:
            ampli_vmin = ampli_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                ampli_vmin, ampli_vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None

        if phase_vmin is None or phase_vmax is None:
            phase_vmin = phase_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                phase_vmin, phase_vmax = (-180.0, 180.0)
        else:
            norm = None
    else:
        if ampli_vmin is None:
            ampli_vmin = np.ma.min(ampli_field_2D)
        if ampli_vmax is None:
            ampli_vmax = np.ma.max(ampli_field_2D)
        if phase_vmin is None:
            phase_vmin = np.ma.min(phase_field_2D)
        if phase_vmax is None:
            phase_vmax = np.ma.max(phase_field_2D)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_angle_Doppler_title(
            spectra, field_name, ang, ind_rng, along_azi=along_azi, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolor(
        xaxis_lim,
        yaxis_lim,
        ampli_field_2D,
        cmap=cmap,
        vmin=ampli_vmin,
        vmax=ampli_vmax,
        norm=norm,
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Amplitude")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Amplitude (-)")

    ax = fig.add_subplot(122)
    cax = ax.pcolor(
        xaxis_lim,
        yaxis_lim,
        phase_field_2D,
        cmap=cmap,
        vmin=phase_vmin,
        vmax=phase_vmax,
        norm=norm,
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Phase")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Phase (deg)")

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_amp_phase_time_Doppler(
    spectra,
    field_name,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    yaxis_pos="start",
    titl=None,
    ampli_vmin=None,
    ampli_vmax=None,
    phase_vmin=None,
    phase_vmax=None,
):
    """
    Makes a complex time-Doppler plot plotting separately the module and the
    phase of the signal

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'
    titl : str or None
        The plot title
    ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"]
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"]
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.ma.masked_all((spectra.nrays, spectra.npulses_max))
        for ray, npuls in enumerate(spectra.npulses["data"]):
            xaxis[ray, 0:npuls] = np.arange(npuls)
        xlabel = "Pulse number"
    ylabel = "time (s from start time)"

    xaxis_lim, yaxis_lim, xaxis_size = _create_irregular_grid(
        xaxis, spectra.time["data"], yaxis_pos=yaxis_pos
    )

    field_2D = _adapt_data_to_irregular_grid(
        np.ma.squeeze(spectra.fields[field_name]["data"], axis=1),
        xaxis_size,
        xaxis_lim.shape[0],
        xaxis_lim.shape[1],
    )

    ampli_field_2D = np.ma.abs(field_2D)
    phase_field_2D = np.ma.angle(field_2D, deg=True)

    # display data
    norm = None
    cmap = None
    ticks = None
    ticklabs = None
    if field_name is not None:
        cmap = pyart.config.get_field_colormap(field_name)

        norm, ticks, ticklabs = get_norm(
            field_name, field_dict=spectra.fields[field_name]
        )
        if ampli_vmin is None or ampli_vmax is None:
            ampli_vmin = ampli_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                ampli_vmin, ampli_vmax = pyart.config.get_field_limits(field_name)
        else:
            norm = None

        if phase_vmin is None or phase_vmax is None:
            phase_vmin = phase_vmax = None
            if norm is None:  # if norm is set do not override with vmin/vmax
                phase_vmin, phase_vmax = (-180.0, 180.0)
        else:
            norm = None
    else:
        if ampli_vmin is None:
            ampli_vmin = np.ma.min(ampli_field_2D)
        if ampli_vmax is None:
            ampli_vmax = np.ma.max(ampli_field_2D)
        if phase_vmin is None:
            phase_vmin = np.ma.min(phase_field_2D)
        if phase_vmax is None:
            phase_vmax = np.ma.max(phase_field_2D)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, 0, 0, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    cax = ax.pcolor(
        xaxis_lim,
        yaxis_lim,
        ampli_field_2D,
        cmap=cmap,
        vmin=ampli_vmin,
        vmax=ampli_vmax,
        norm=norm,
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Amplitude")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Amplitude (-)")

    ax = fig.add_subplot(122)
    cax = ax.pcolor(
        xaxis_lim,
        yaxis_lim,
        phase_field_2D,
        cmap=cmap,
        vmin=phase_vmin,
        vmax=phase_vmax,
        norm=norm,
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Phase")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    cb = fig.colorbar(cax)
    if ticks is not None:
        cb.set_ticks(ticks)
    if ticklabs:
        cb.set_ticklabels(ticklabs)
    cb.set_label("Phase (deg)")

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_complex_Doppler(
    spectra,
    field_name,
    ray,
    rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    ylabel=None,
    titl=None,
    vmin=None,
    vmax=None,
):
    """
    Makes a complex Doppler plot plotting separately the real and the
    imaginary parts

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray, rng : int
        ray and range index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    ylabel : str or None
        The label of the y-axis
    titl : str or None
        The plot title
    vmin, vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"

    field = spectra.fields[field_name]["data"][ray, rng, 0 : xaxis.size]

    re_field = np.real(field)
    im_field = np.imag(field)

    # display data
    if field_name is not None:
        if ylabel is None:
            field_dict = pyart.config.get_metadata(field_name)
            ylabel = get_colobar_label(field_dict, field_name)
        if vmin is None or vmax is None:
            vmin, vmax = pyart.config.get_field_limits(field_name)
    else:
        if vmin is None:
            vmin = np.ma.min([np.ma.min(re_field), np.ma.min(im_field)])
        if vmax is None:
            vmax = np.ma.max([np.ma.max(re_field), np.ma.max(im_field)])

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, ray, rng, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    ax.plot(xaxis, re_field, marker="x")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_ylim(bottom=vmin, top=vmax)
    ax.set_xlim([xaxis[0], xaxis[-1]])
    ax.set_title("Real part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    # Turn on the grid
    ax.grid()

    ax = fig.add_subplot(122)
    ax.plot(xaxis, im_field, marker="x")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_ylim(bottom=vmin, top=vmax)
    ax.set_xlim([xaxis[0], xaxis[-1]])
    ax.set_title("Imaginary part")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    # Turn on the grid
    ax.grid()

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_amp_phase_Doppler(
    spectra,
    field_name,
    ray,
    rng,
    prdcfg,
    fname_list,
    xaxis_info="Doppler_velocity",
    titl=None,
    ampli_vmin=None,
    ampli_vmax=None,
    phase_vmin=None,
    phase_vmax=None,
):
    """
    Makes a complex Doppler plot plotting separately the module and the phase
    of the signal

    Parameters
    ----------
    spectra : radar spectra object
        object containing the spectra or the IQ data to plot
    field_name : str
        name of the field to plot
    ray, rng : int
        ray and range index
    prdcfg : dict
        dictionary containing the product configuration
    fname_list : list of str
        list of names of the files where to store the plot
    xaxis_info : str
        Type of x-axis. Can be 'Doppler_velocity', 'Doppler_frequency' or
        'pulse_number'
    titl : str or None
        The plot title
    ampli_vmin, ampli_vmax, phase_vmin, phase_vmax : float or None
        The value limits

    Returns
    -------
    fname_list : list of str
        list of names of the saved plots

    """
    dpi = prdcfg["spectraImageConfig"].get("dpi", 72)
    xsize = prdcfg["spectraImageConfig"]["xsize"]
    ysize = prdcfg["spectraImageConfig"]["ysize"]
    velmin = prdcfg["spectraImageConfig"]["velmin"]
    velmax = prdcfg["spectraImageConfig"]["velmax"]

    if xaxis_info == "Doppler_velocity":
        xaxis = spectra.Doppler_velocity["data"][ray, :].compressed()
        xlabel = "Doppler velocity (m/s)"
    elif xaxis_info == "Doppler_frequency":
        xaxis = spectra.Doppler_frequency["data"][ray, :].compressed()
        xlabel = "Doppler frequency (Hz)"
    else:
        xaxis = np.arange(spectra.npulses["data"][ray])
        xlabel = "Pulse number"

    field = spectra.fields[field_name]["data"][ray, rng, 0 : xaxis.size]

    ampli_field = np.ma.abs(field)
    phase_field = np.ma.angle(field, deg=True)

    # display data
    if field_name is not None:
        if ampli_vmin is None or ampli_vmax is None:
            ampli_vmin, ampli_vmax = pyart.config.get_field_limits(field_name)
        if phase_vmin is None or phase_vmax is None:
            phase_vmin, phase_vmax = (-180.0, 180.0)
    else:
        if ampli_vmin is None:
            ampli_vmin = np.ma.min(ampli_field)
        if ampli_vmax is None:
            ampli_vmax = np.ma.max(ampli_field)
        if phase_vmin is None:
            phase_vmin = np.ma.min(phase_field)
        if phase_vmax is None:
            phase_vmax = np.ma.max(phase_field)

    fig = plt.figure(figsize=[xsize, ysize], dpi=dpi)
    if titl is None:
        titl = generate_complex_Doppler_title(
            spectra, field_name, ray, rng, datetime_format=None
        )

    plt.suptitle(titl)

    ax = fig.add_subplot(121)
    ax.plot(xaxis, ampli_field, marker="x")
    ax.set_xlabel(xlabel)
    ax.set_ylabel("Amplitude (-)")
    ax.set_ylim(bottom=ampli_vmin, top=ampli_vmax)
    ax.set_xlim([xaxis[0], xaxis[-1]])
    ax.set_title("Amplitude")
    if not (velmin is None or velmax is None):
        ax.set_xlim([velmin, velmax])

    # Turn on the grid
    ax.grid()

    ax = fig.add_subplot(122)
    ax.plot(xaxis, phase_field, marker="x")
    ax.set_xlabel(xlabel)
    ax.set_ylabel("Phase (deg)")
    ax.set_ylim(bottom=phase_vmin, top=phase_vmax)
    ax.set_xlim([xaxis[0], xaxis[-1]])
    ax.set_title("Phase")

    # Turn on the grid
    ax.grid()

    # Make a tight layout
    fig.tight_layout()

    plt.subplots_adjust(top=0.8)

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def _create_irregular_grid(xaxis, yaxis, yaxis_pos="start"):
    """
    Create an irregular grid to be able to plot data with variable x-axis

    Parameters
    ----------
    xaxis : 2D float array
        array containing the x points
    yaxis : 1D float array
        array containing the y points
    yaxis_pos : str
        the position that the y point represents in the y-axis bin. Can be
        'start', end' or 'centre'

    Returns
    -------
    xaxis_lim, yaxis_lim : 2D float array
        Matrix containing the edges of the X and Y axis

    """
    nbinsy = yaxis.size

    # One dimensional y-axis bin limits
    yres = np.median(yaxis[1:] - yaxis[:-1])
    if yaxis_pos == "start":
        yaxis_lim = np.append(yaxis, yaxis[-1] + yres)
    elif yaxis_pos == "end":
        yaxis_lim = np.append(yaxis - yres, yaxis[-1])
    else:
        yaxis_lim = np.append(yaxis - yres / 2.0, yaxis[-1] + yres / 2.0)

    # List of x-axis bin limits
    xres = np.abs(xaxis[:, 1] - xaxis[:, 0])
    xaxis_size = list()
    xaxis_list = list()
    for i in range(nbinsy):
        xaxis_aux = xaxis[i, :].compressed()
        xaxis_lim = np.append(xaxis_aux - xres[i] / 2.0, xaxis_aux[-1] + xres[i] / 2.0)
        xaxis_size.append(xaxis_lim.size)
        xaxis_list.append(xaxis_lim)

    nxbin_lim = max(xaxis_size)

    # Create two dimensional grid of y-axis
    yaxis_lim = np.repeat(yaxis_lim, 2)[1:-1]
    nybin_lim = yaxis_lim.size
    yaxis_lim = np.broadcast_to(
        np.reshape(yaxis_lim, (1, nybin_lim)), (nxbin_lim, nybin_lim)
    )

    # Create two dimensional grid of x-axis
    xaxis_lim = np.ma.masked_all((nxbin_lim, nbinsy))
    for i in range(nbinsy):
        xaxis_lim[0 : xaxis_size[i], i] = xaxis_list[i]
    xaxis_lim = np.repeat(xaxis_lim, 2, axis=1)

    return xaxis_lim, yaxis_lim, xaxis_size


def _adapt_data_to_irregular_grid(data, xaxis_size, nxbin_lim, nybin_lim):
    """
    Adapts data to irregular grid to allow plotting

    Parameters
    ----------
    data : 2D float array
        The data to plot
    xaxis_size : 1D-array
        The size of each x-axis
    nxbin_lim, nybin_lim : float
        number of gate limits at each axis

    Returns
    -------
    zpoints : 2D float array
        Matrix containing the data points.

    """
    nbinsy = data.shape[0]

    # adapt Z data
    zpoints = np.ma.masked_all((nxbin_lim - 1, nybin_lim - 1), dtype=data.dtype)
    for i in range(nbinsy - 1):
        zpoints[0 : xaxis_size[i] - 1, 2 * i] = data[i, 0 : xaxis_size[i] - 1]
        zpoints[0 : xaxis_size[i] - 1, 2 * i + 1] = data[i, 0 : xaxis_size[i] - 1]
    zpoints[0 : xaxis_size[-1] - 1, -1] = data[-1, 0 : xaxis_size[-1] - 1]

    return zpoints



"""
pyrad.graph.plot_timeseries
===========================

Functions to plot Pyrad datasets

.. autosummary::
    :toctree: generated/

    plot_timeseries
    plot_timeseries_comp
    plot_monitoring_ts
    plot_intercomp_scores_ts
    plot_ml_ts
    plot_sun_retrieval_ts

"""

import pyart
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from warnings import warn

import numpy as np
import datetime

import matplotlib as mpl

mpl.use("Agg")

# Increase a bit font size
mpl.rcParams.update({"font.size": 16})
mpl.rcParams.update({"font.family": "sans-serif"})


def plot_timeseries(
    tvec,
    data_list,
    fname_list,
    labelx="Time [UTC]",
    labely="Value",
    labels=["Sensor"],
    title="Time Series",
    period=0,
    timeformat=None,
    colors=None,
    linestyles=None,
    markers=None,
    ymin=None,
    ymax=None,
    dpi=72,
):
    """
    plots a time series

    Parameters
    ----------
    tvec : datetime object
        time of the time series
    data_list : list of float array
        values of the time series
    fname_list : list of str
        list of names of the files where to store the plot
    labelx : str
        The label of the X axis
    labely : str
        The label of the Y axis
    labels : array of str
        The label of the legend
    title : str
        The figure title
    period : float
        measurement period in seconds used to compute accumulation. If 0 no
        accumulation is computed
    timeformat : str
        Specifies the tvec and time format on the x axis
    colors : array of str
        Specifies the colors of each line
    linestyles : array of str
        Specifies the line style of each line
    markers: array of str
        Specify the markers to be used for each line
    ymin, ymax: float
        Lower/Upper limit of y axis
    dpi : int
        dots per inch

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    History
    --------
    201?.??.?? -fvj- creation
    2017.08.21 -jgr- modified margins and grid + minor graphical updates
    2018.03.05 -jgr- added x-limit of x axis to avoid unwanted error messages

    """
    tvec = np.array(tvec)  # convert  from pandas if needed
    if period > 0:
        for i, data in enumerate(data_list):
            data *= period / 3600.0
            data_list[i] = np.ma.cumsum(data)

    fig, ax = plt.subplots(figsize=[10, 6], dpi=dpi)

    lab = None
    col = None
    lstyle = "--"
    marker = "o"

    for i, data in enumerate(data_list):
        if labels is not None:
            lab = labels[i]
        if colors is not None:
            col = colors[i]
        if linestyles is not None:
            lstyle = linestyles[i]
        if markers is not None:
            marker = markers[i]
        ax.plot(tvec, data, label=lab, color=col, linestyle=lstyle, marker=marker)

    ax.set_title(title)
    ax.set_xlabel(labelx)
    ax.set_ylabel(labely)
    ax.set_ylim(bottom=ymin, top=ymax)
    ax.set_xlim([tvec[0], tvec[-1]])

    # Turn on the grid
    ax.grid()

    if timeformat is not None:
        ax.xaxis.set_major_formatter(mdates.DateFormatter(timeformat))

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_timeseries_comp(
    date1,
    value1,
    date2,
    value2,
    fname_list,
    labelx="Time [UTC]",
    labely="Value",
    label1="Sensor 1",
    label2="Sensor 2",
    titl="Time Series Comparison",
    period1=0,
    period2=0,
    ymin=None,
    ymax=None,
    dpi=72,
):
    """
    plots 2 time series in the same graph

    Parameters
    ----------
    date1 : datetime object
        time of the first time series
    value1 : float array
        values of the first time series
    date2 : datetime object
        time of the second time series
    value2 : float array
        values of the second time series
    fname_list : list of str
        list of names of the files where to store the plot
    labelx : str
        The label of the X axis
    labely : str
        The label of the Y axis
    label1, label2 : str
        legend label for each time series
    titl : str
        The figure title
     period1, period2 : float
        measurement period in seconds used to compute accumulation. If 0 no
        accumulation is computed
    dpi : int
        dots per inch
    ymin, ymax : float
        The limits of the Y-axis. None will keep the default limit.

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    History
    --------
    201?.??.?? -fvj- created
    2017.08.21 -jgr- changed some graphical aspects

    """
    if (period1 > 0) and (period2 > 0):
        # TODO: document this and check (sometimes artefacts)
        value1 *= period1 / 3600.0
        value1 = np.ma.cumsum(value1)

        value2 *= period2 / 3600.0
        value2 = np.ma.cumsum(value2)

    fig, ax = plt.subplots(figsize=[10, 6.5], dpi=dpi)
    ax.plot(date1, value1, "b", label=label1, linestyle="--", marker="o")
    ax.plot(date2, value2, "r", label=label2, linestyle="--", marker="s")
    ax.legend(loc="best")
    ax.set_xlabel(labelx)
    ax.set_ylabel(labely)
    ax.set_title(titl)

    ax.grid()

    ax.set_ylim(bottom=ymin, top=ymax)
    ax.set_xlim([date2[0], date2[-1]])

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()

    # Make a tight layout
    fig.tight_layout()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_monitoring_ts(
    date,
    np_t,
    val_mean_and_quant,
    quantiles,
    field_name,
    fname_list,
    ref_value=None,
    vmin=None,
    vmax=None,
    np_min=0,
    labelx="Time [UTC]",
    labely="Value",
    titl="Time Series",
    dpi=72,
    plot_until_year_end=False,
):
    """
    Plots a time series of monitoring data.

    Parameters
    ----------
    date : datetime object
        Time of the time series.
    np_t : int array
        Number of points.
    val_mean_and_quant : float array
        Values of the mean in dB, mean in dB computed from linear units,
        followed by all quantiles.
    quantiles : 1D ndarray
        The computed quantiles.
    field_name : str
        Name of the field.
    fname_list : list of str
        List of filenames where to store the plot.
    ref_value : float, optional
        The reference value.
    vmin, vmax : float, optional
        Limits of the y-axis.
    np_min : int, optional
        Minimum number of points to consider the sample plottable.
    labelx : str, optional
        Label for the X-axis.
    labely : str, optional
        Label for the Y-axis.
    titl : str, optional
        The figure title.
    dpi : int, optional
        Dots per inch.
    plot_until_year_end : bool, optional
        If True, sets the maximum x-tick to the end of the year.

    Returns
    -------
    fname_list : list of str
        List of filenames of the created plots.
    """

    # Get default Py-ART field limits
    vmin_pyart, vmax_pyart = pyart.config.get_field_limits(field_name)
    if vmin is None:
        vmin = vmin_pyart
    if vmax is None:
        vmax = vmax_pyart

    date = np.array(date)

    # Separate mean values and quantiles from val_mean_and_quant
    geometric_mean_dB = val_mean_and_quant[:, 0]
    linear_mean_dB = val_mean_and_quant[:, 1]
    quantile_values = val_mean_and_quant[:, 2:]

    # Define colors for plotting
    quantile_colors = plt.cm.viridis_r(np.linspace(0, 1, len(quantiles)))

    # Determine valid data points
    isvalid = np.logical_not(np.ma.getmaskarray(geometric_mean_dB))
    if np_min > 0:
        has_np = np_t > np_min
        isvalid = np.logical_and(isvalid, has_np)

    date_plt = date[isvalid]
    mean_dB_plt = geometric_mean_dB[isvalid]
    mean_dB_from_lin_plt = linear_mean_dB[isvalid]
    quantile_values_plt = quantile_values[isvalid]

    fig = plt.figure(figsize=[15, 13], dpi=dpi)

    ax = fig.add_subplot(2, 1, 1)

    # Plot means
    ax.plot(date_plt, mean_dB_plt, "rx--", label="Geometric mean")
    ax.plot(date_plt, mean_dB_from_lin_plt, "bx--", label="Linear mean")

    # Plot all quantiles dynamically
    for i, q in enumerate(quantiles):
        ax.plot(
            date_plt,
            quantile_values_plt[:, i],
            "x-",
            color=quantile_colors[i],
            label=f"Q{q}",
        )

    # Plot reference value if provided
    if ref_value is not None:
        ax.axhline(ref_value, color="k", linestyle="--", label="Reference")

    ax.set_ylabel(labely)
    ax.set_title(titl)
    ax.set_ylim([vmin, vmax])
    ax.grid(True)
    ax.legend(loc="best")

    # Adjust x-axis limits
    if plot_until_year_end:
        t0 = date_plt[0]
        tend = datetime.datetime(year=t0.year, month=12, day=31, hour=23, minute=59)
        ax.set_xlim([t0, tend])
    else:
        ax.autoscale(enable=True, axis="x", tight=True)
        ax.set_xlim([date_plt[0], date_plt[-1]])

    # Second subplot for NP values
    ax = fig.add_subplot(2, 1, 2)
    ax.plot(date, np_t, "x-", label="Number of Samples")

    if np_min is not None:
        ax.axhline(np_min, color="k", linestyle="--", label=f"Min NP ({np_min})")

    if plot_until_year_end:
        ax.set_xlim([t0, tend])
    else:
        ax.autoscale(enable=True, axis="x", tight=True)
        ax.set_xlim([date_plt[0], date_plt[-1]])

    ax.set_ylabel("Number of Samples")
    ax.set_xlabel(labelx)
    ax.legend(loc="best")
    fig.autofmt_xdate()

    plt.tight_layout()

    # Save figure
    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_intercomp_scores_ts(
    date_vec,
    np_vec,
    meanbias_vec,
    medianbias_vec,
    quant25bias_vec,
    quant75bias_vec,
    modebias_vec,
    corr_vec,
    slope_vec,
    intercep_vec,
    intercep_slope1_vec,
    fname_list,
    ref_value=0.0,
    np_min=0,
    corr_min=0.0,
    labelx="Time UTC",
    titl="RADAR001-RADAR002 intercomparison",
    dpi=72,
):
    """
    plots a time series of radar intercomparison scores

    Parameters
    ----------
    date_vec : datetime object
        time of the time series
    np_vec : int array
        number of points
    meanbias_vec, medianbias_vec, modebias_vec : float array
        mean, median and mode bias
    quant25bias_vec, quant75bias_vec: 25th and 75th percentile of the bias
    corr_vec : float array
        correlation
    slope_vec, intercep_vec : float array
        slope and intercep of a linear regression
    intercep_slope1_vec : float
        the intercep point of a inear regression of slope 1
    ref_value : float
        the reference value
    np_min : int
        The minimum number of points to consider the result valid
    corr_min : float
        The minimum correlation to consider the results valid
    labelx : str
        The label of the X axis
    titl : str
        The figure title

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    # plot only valid data (but keep first and last date)
    date2 = np.array(date_vec)
    isvalid = np.logical_not(np.ma.getmaskarray(meanbias_vec))
    isvalid_corr = np.logical_not(np.ma.getmaskarray(corr_vec))
    if np_min > 0:
        has_np = np_vec > np_min
        isvalid = np.logical_and(isvalid, has_np)
    if corr_min > 0:
        has_corr_min = corr_vec > corr_min
        isvalid = np.logical_and(isvalid, has_corr_min)

    meanbias_plt = meanbias_vec[isvalid]
    medianbias_plt = medianbias_vec[isvalid]
    quant25bias_plt = quant25bias_vec[isvalid]
    quant75bias_plt = quant75bias_vec[isvalid]
    modebias_plt = modebias_vec[isvalid]
    intercep_plt = intercep_slope1_vec[isvalid]
    corr_plt = corr_vec[isvalid_corr]
    date_corr = date2[isvalid_corr]
    date_plt = date2[isvalid]
    if not isvalid[0]:
        meanbias_plt = np.ma.append(np.ma.masked, meanbias_plt)
        medianbias_plt = np.ma.append(np.ma.masked, medianbias_plt)
        quant25bias_plt = np.ma.append(np.ma.masked, quant25bias_plt)
        quant75bias_plt = np.ma.append(np.ma.masked, quant75bias_plt)
        modebias_plt = np.ma.append(np.ma.masked, modebias_plt)
        intercep_plt = np.ma.append(np.ma.masked, intercep_plt)
        date_plt = np.ma.append(date2[0], date_plt)
    if not isvalid[-1]:
        meanbias_plt = np.ma.append(meanbias_plt, np.ma.masked)
        medianbias_plt = np.ma.append(medianbias_plt, np.ma.masked)
        quant25bias_plt = np.ma.append(quant25bias_plt, np.ma.masked)
        quant75bias_plt = np.ma.append(quant75bias_plt, np.ma.masked)
        modebias_plt = np.ma.append(modebias_plt, np.ma.masked)
        intercep_plt = np.ma.append(intercep_plt, np.ma.masked)
        date_plt = np.ma.append(date_plt, date2[-1])

    if not isvalid_corr[0]:
        corr_plt = np.ma.append(np.ma.masked, corr_plt)
        date_corr = np.ma.append(date2[0], date_corr)
    if not isvalid_corr[-1]:
        corr_plt = np.ma.append(corr_plt, np.ma.masked)
        date_corr = np.ma.append(date_corr, date2[-1])

    fig = plt.figure(figsize=[10, 20], dpi=dpi)

    ax = fig.add_subplot(4, 1, 1)
    ax.plot(date_plt, medianbias_plt, "bx-", label="median")
    ax.plot(date_plt, meanbias_plt, "rx-", label="mean")
    ax.plot(date_plt, modebias_plt, "gx-", label="mode")
    ax.plot(date_plt, intercep_plt, "yx-", label="intercep of slope 1 LR")
    if ref_value is not None:
        ax.plot(date_plt, np.zeros(len(date_plt)) + ref_value, "k--")
    # plt.legend(loc='best')
    ax.set_ylabel("bias [dB]")
    ax.set_title(titl)
    ax.set_ylim([-5.0, 5.0])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    ax = fig.add_subplot(4, 1, 2)
    ax.plot(date_plt, medianbias_plt, "bx-", label="median")
    ax.plot(date_plt, quant25bias_plt, "rx-", label="25-percentile")
    ax.plot(date_plt, quant75bias_plt, "rx-", label="75-percentile")
    if ref_value is not None:
        ax.plot(date_plt, np.zeros(len(date_plt)) + ref_value, "k--")
    # plt.legend(loc='best')
    ax.set_ylabel("bias [dB]")
    ax.set_ylim([-5.0, 5.0])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    ax = fig.add_subplot(4, 1, 3)
    ax.plot(date_corr, corr_plt, "bx-")

    if corr_min > 0:
        ax.plot(date_corr, np.zeros(len(date_corr)) + corr_min, "k--")

    ax.set_ylabel("correlation")
    ax.set_ylim([0.0, 1.0])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    ax = fig.add_subplot(4, 1, 4)
    ax.plot(date2, np_vec, "bx-")

    if np_min > 0:
        ax.plot(date2, np.zeros(len(date2)) + np_min, "k--")

    ax.set_ylabel("Number of Samples")
    ax.set_xlabel(labelx)

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)
    return fname_list


def plot_ml_ts(
    dt_ml_arr,
    ml_top_avg_arr,
    ml_top_std_arr,
    thick_avg_arr,
    thick_std_arr,
    nrays_valid_arr,
    nrays_total_arr,
    fname_list,
    labelx="Time UTC",
    titl="Melting layer time series",
    dpi=72,
):
    """
    plots a time series of melting layer data

    Parameters
    ----------
    dt_ml_arr : datetime object
        time of the time series
    np_vec : int array
        number of points
    meanbias_vec, medianbias_vec, modebias_vec : float array
        mean, median and mode bias
    quant25bias_vec, quant75bias_vec: 25th and 75th percentile of the bias
    corr_vec : float array
        correlation
    slope_vec, intercep_vec : float array
        slope and intercep of a linear regression
    intercep_slope1_vec : float
        the intercep point of a inear regression of slope 1
    ref_value : float
        the reference value
    np_min : int
        The minimum number of points to consider the result valid
    corr_min : float
        The minimum correlation to consider the results valid
    labelx : str
        The label of the X axis
    titl : str
        The figure title

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    fig = plt.figure(figsize=[10, 15], dpi=dpi)

    ax = fig.add_subplot(3, 1, 1)
    ax.plot(dt_ml_arr, ml_top_avg_arr, "bx-", label="avg")
    ax.plot(dt_ml_arr, ml_top_avg_arr + ml_top_std_arr, "rx-", label="avg+std")
    ax.plot(dt_ml_arr, ml_top_avg_arr - ml_top_std_arr, "rx-", label="avg-std")
    # plt.legend(loc='best')
    ax.set_ylabel("Top height [m MSL]")
    ax.set_title(titl)
    ax.set_ylim([0.0, 6000.0])
    ax.set_xlim([dt_ml_arr[0], dt_ml_arr[-1]])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    ax = fig.add_subplot(3, 1, 2)
    ax.plot(dt_ml_arr, thick_avg_arr, "bx-", label="avg")
    ax.plot(dt_ml_arr, thick_avg_arr + thick_std_arr, "rx-", label="avg+std")
    ax.plot(dt_ml_arr, thick_avg_arr - thick_std_arr, "rx-", label="avg-std")
    # plt.legend(loc='best')
    ax.set_ylabel("Thickness [m]")
    ax.set_ylim([0.0, 3000.0])
    ax.set_xlim([dt_ml_arr[0], dt_ml_arr[-1]])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    ax = fig.add_subplot(3, 1, 3)
    ax.plot(dt_ml_arr, nrays_valid_arr, "bx-", label="N valid rays")
    ax.plot(dt_ml_arr, nrays_total_arr, "rx-", label="rays total")
    # plt.legend(loc='best')
    ax.set_ylabel("Rays")
    ax.set_xlabel(labelx)
    ax.set_ylim([0, np.max(nrays_total_arr) + 5])
    ax.set_xlim([dt_ml_arr[0], dt_ml_arr[-1]])

    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list


def plot_sun_retrieval_ts(
    sun_retrieval,
    data_type,
    fname_list,
    labelx="Date",
    titl="Sun retrieval Time Series",
    dpi=72,
):
    """
    plots sun retrieval time series series

    Parameters
    ----------
    sun_retrieval : tuple
        tuple containing the retrieved parameters
    data_type : str
        parameter to be plotted
    fname_list : list of str
        list of names of the files where to store the plot
    labelx : str
        the x label
    titl : str
        the title of the plot
    dpi : int
        dots per inch

    Returns
    -------
    fname_list : list of str
        list of names of the created plots

    """
    value_std = None
    ref = None
    date = sun_retrieval[1]
    if data_type == "nhits_h":
        value = sun_retrieval[2]
        labely = "Number of sun hits H channel"
        vmin = 0
        vmax = np.max(sun_retrieval[2]) + 1
    elif data_type == "el_width_h":
        value = sun_retrieval[3]
        labely = "Elevation beamwidth H channel (Deg)"
        vmin = 0.0
        vmax = 4.0
    elif data_type == "az_width_h":
        value = sun_retrieval[4]
        labely = "Azimuth beamwidth H channel (Deg)"
        vmin = 0.0
        vmax = 4.0
    elif data_type == "el_bias_h":
        value = sun_retrieval[5]
        ref = np.zeros(len(value))
        labely = "Elevation pointing bias H channel (Deg)"
        vmin = -2.0
        vmax = 2.0
    elif data_type == "az_bias_h":
        value = sun_retrieval[6]
        ref = np.zeros(len(value))
        labely = "Azimuth pointing bias H channel (Deg)"
        vmin = -2.0
        vmax = 2.0
    elif data_type == "dBm_sun_est":
        value = sun_retrieval[7]
        value_std = sun_retrieval[8]
        labely = "Sun Power H channel (dBm)"
        vmin = -110.0
        vmax = -90.0
    elif data_type == "rx_bias_h":
        value = 10.0 * np.ma.log10(sun_retrieval[9]) - 10.0 * np.ma.log10(
            sun_retrieval[21]
        )
        value_std = sun_retrieval[8]
        ref = np.zeros(len(value))
        labely = "Receiver bias H channel (dB)"
        vmin = -5.0
        vmax = 5.0
    elif data_type == "sf_h":
        value = 10.0 * np.ma.log10(sun_retrieval[9])
        # value_std = sun_retrieval[8]
        ref = 10.0 * np.ma.log10(sun_retrieval[21])
        labely = "Observed solar flux H channel (dB(sfu))"
        vmin = 15.0
        vmax = 30.0
    elif data_type == "nhits_v":
        value = sun_retrieval[10]
        labely = "Number of sun hits V channel"
        vmin = 0
        vmax = np.max(sun_retrieval[10]) + 1
    elif data_type == "el_width_v":
        value = sun_retrieval[11]
        labely = "Elevation beamwidth V channel (Deg)"
        vmin = 0.0
        vmax = 4.0
    elif data_type == "az_width_v":
        value = sun_retrieval[12]
        labely = "Azimuth beamwidth V channel (Deg)"
        vmin = 0.0
        vmax = 4.0
    elif data_type == "el_bias_v":
        value = sun_retrieval[13]
        ref = np.zeros(len(value))
        labely = "Elevation pointing bias V channel (Deg)"
        vmin = -2.0
        vmax = 2.0
    elif data_type == "az_bias_v":
        value = sun_retrieval[14]
        ref = np.zeros(len(value))
        labely = "Azimuth pointing bias V channel (Deg)"
        vmin = -2.0
        vmax = 2.0
    elif data_type == "dBmv_sun_est":
        value = sun_retrieval[15]
        value_std = sun_retrieval[16]
        labely = "Sun Power V channel (dBm)"
        vmin = -110.0
        vmax = -90.0
    elif data_type == "rx_bias_v":
        value = 10.0 * np.ma.log10(sun_retrieval[17]) - 10.0 * np.ma.log10(
            sun_retrieval[21]
        )
        value_std = sun_retrieval[16]
        ref = np.zeros(len(value))
        labely = "Receiver bias V channel (dB)"
        vmin = -5.0
        vmax = 5.0
    elif data_type == "sf_v":
        value = 10.0 * np.ma.log10(sun_retrieval[17])
        # value_std = sun_retrieval[16]
        ref = 10.0 * np.ma.log10(sun_retrieval[21])
        labely = "Observed solar flux V channel (dB(sfu))"
        vmin = 15.0
        vmax = 30.0
    elif data_type == "nhits_zdr":
        value = sun_retrieval[18]
        labely = "Number of sun hits ZDR"
        vmin = 0
        vmax = np.max(sun_retrieval[18]) + 1
    elif data_type == "ZDR_sun_est":
        value = sun_retrieval[19]
        value_std = sun_retrieval[20]
        ref = np.zeros(len(value))
        labely = "Sun ZDR (dB)"
        vmin = -2.0
        vmax = 2.0

    mask = np.ma.getmaskarray(value)
    if mask.all():
        warn("Unable to create figure " + " ".join(fname_list) + ". No valid data")
        return None

    # plot only valid data (but keep first and last date)
    isvalid = np.logical_not(mask)
    date2 = np.array(date)

    value_plt = value[isvalid]
    date_plt = date2[isvalid]
    if not isvalid[0]:
        value_plt = np.ma.append(np.ma.masked, value_plt)
        date_plt = np.ma.append(date2[0], date_plt)
    if not isvalid[-1]:
        value_plt = np.ma.append(value_plt, np.ma.masked)
        date_plt = np.ma.append(date_plt, date2[-1])

    fig, ax = plt.subplots(figsize=[10, 6], dpi=dpi)
    ax.plot(date_plt, value_plt, "x-")
    if value_std is not None:
        value_std_plt = value_std[isvalid]
        if not isvalid[0]:
            value_std_plt = np.ma.append(np.ma.masked, value_std_plt)
        if not isvalid[-1]:
            value_std_plt = np.ma.append(value_std_plt, np.ma.masked)

        ax.plot(date_plt, value_plt + value_std_plt, "rx-")
        ax.plot(date_plt, value_plt - value_std_plt, "rx-")
    if ref is not None:
        ref_plt = ref[isvalid]
        if not isvalid[0]:
            ref_plt = np.ma.append(ref[0], ref_plt)
        if not isvalid[-1]:
            ref_plt = np.ma.append(ref_plt, ref[-1])
        ax.plot(date_plt, ref_plt, "k--")
    ax.set_xlabel(labelx)
    ax.set_ylabel(labely)
    ax.set_title(titl)
    ax.set_ylim([vmin, vmax])
    ax.set_xlim([date_plt[0], date_plt[-1]])
    # tight x axis
    ax.autoscale(enable=True, axis="x", tight=True)
    ax.grid(True)

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()

    for fname in fname_list:
        fig.savefig(fname, dpi=dpi)
    plt.close(fig)

    return fname_list



"""
pyrad.proc.process_Doppler
===========================

Functions for processing Doppler related parameters

.. autosummary::
    :toctree: generated/

    process_turbulence
    process_dealias_fourdd
    process_dealias_region_based
    process_dealias_unwrap_phase
    process_radial_velocity
    process_wind_vel
    process_windshear
    process_windshear_lidar
    process_vad
    process_dda

"""

from copy import deepcopy
from warnings import warn
import numpy as np

import pyart


try:
    import pytda

    _PYTDA_AVAILABLE = True
except ImportError:
    _PYTDA_AVAILABLE = False

try:
    import pydda

    _PYDDA_AVAILABLE = True
except ImportError:
    _PYDDA_AVAILABLE = False

from ..io.io_aux import (
    get_datatype_fields,
    get_fieldname_pyart,
    convert_pydda_to_pyart_grid,
)
from ..io import read_radiosounding_igra
from .process_grid import process_grid
from ..util import compute_average_vad


def process_turbulence(procstatus, dscfg, radar_list=None):
    """
    Computes turbulence from the Doppler spectrum width and reflectivity using
    the PyTDA package

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain,
            "dBuZ" or "dBZ" or "dBZc" or "dBuZv" or "dBZv" or "dBZvc" or "CNRc", and,
            "W" or "Wv" or "Wu" or "Wvu" or "WD" or "WDc"
        radius : float. Dataset keyword
            Search radius for calculating Eddy Dissipation Rate (EDR).
            Default 2
        split_cut : Bool. Dataset keyword
            Set to True for split-cut volumes. Default False
        max_split_cut : Int. Dataset keyword
            Total number of tilts that are affected by split cuts. Only
            relevant if split_cut=True. Default 2
        xran, yran : float array. Dataset keyword
            Spatial range in X,Y to consider. Default [-100, 100] for both
            X and Y
        use_ntda : Bool. Dataset keyword
            Wether to use NCAR Turbulence Detection Algorithm (NTDA). Default
            True
        beamwidth : Float. Dataset keyword
            Radar beamwidth. Default None. If None it will be obtained from
            the radar object metadata. If cannot be obtained defaults to 1
            deg.
        compute_gate_pos : Bool. Dataset keyword
            If True the gate position is going to be computed in PyTDA.
            Otherwise the position from the radar object is used. Default
            False
        verbose : Bool. Dataset keyword
            True for verbose output. Default False

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "EDR"
    ind_rad : int
        radar index

    """
    if not _PYTDA_AVAILABLE:
        warn("PyTDA package not available. Unable to compute turbulence")
        return None, None

    if procstatus != 1:
        return None, None

    width_field = None
    refl_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("dBuZ", "dBZ", "dBZc", "dBuZv", "dBZv", "dBZvc", "CNRc"):
            refl_field = get_fieldname_pyart(datatype)
        if datatype in ("W", "Wv", "Wu", "Wvu", "WD", "WDc"):
            width_field = get_fieldname_pyart(datatype)

    if width_field is None or refl_field is None:
        warn(
            "Reflectivity and spectrum width fields required" " to estimate turbulence"
        )
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if width_field not in radar.fields or refl_field not in radar.fields:
        warn("Unable to compute turbulence. Missing data")
        return None, None

    # user defined parameters
    radius = dscfg.get("radius", 2.0)
    split_cut = dscfg.get("split_cut", False)
    xran = dscfg.get("xran", [-100.0, 100.0])
    yran = dscfg.get("yran", [-100.0, 100.0])
    max_split_cut = dscfg.get("max_split_cut", 2)
    use_ntda = dscfg.get("use_ntda", True)
    beamwidth = dscfg.get("beamwidth", None)
    verbose = dscfg.get("verbose", False)
    compute_gate_pos = dscfg.get("compute_gate_pos", False)

    if beamwidth is None:
        if (
            radar.instrument_parameters is not None
            and "radar_beam_width_h" in radar.instrument_parameters
        ):
            beamwidth = radar.instrument_parameters["radar_beam_width_h"]["data"][0]
        else:
            warn("Unknown radar beamwidth. Default 1 deg will be used")
            beamwidth = 1

    rng_res = (radar.range["data"][1] - radar.range["data"][0]) / 1000.0

    radar_out = deepcopy(radar)
    radar_out.fields = dict()
    radar_out.add_field(refl_field, deepcopy(radar.fields[refl_field]))
    radar_out.add_field(width_field, deepcopy(radar.fields[width_field]))
    radar_out.fields[refl_field]["data"][
        np.ma.getmaskarray(radar_out.fields[refl_field]["data"])
    ] = -32768
    radar_out.fields[width_field]["data"][
        np.ma.getmaskarray(radar_out.fields[width_field]["data"])
    ] = -32768

    radar_out.fields[refl_field]["_FillValue"] = -32768
    radar_out.fields[width_field]["_FillValue"] = -32768

    if radar_out.scan_type == "ppi":
        pytda.calc_turb_vol(
            radar_out,
            radius=radius,
            split_cut=split_cut,
            xran=xran,
            yran=yran,
            verbose=verbose,
            name_dz=refl_field,
            name_sw=width_field,
            turb_name="turbulence",
            max_split_cut=max_split_cut,
            use_ntda=use_ntda,
            beamwidth=beamwidth,
            gate_spacing=rng_res,
            compute_gate_pos=compute_gate_pos,
        )
    elif radar_out.scan_type == "rhi":
        pytda.calc_turb_rhi(
            radar_out,
            radius=radius,
            verbose=verbose,
            name_dz=refl_field,
            name_sw=width_field,
            turb_name="turbulence",
            use_ntda=use_ntda,
            beamwidth=beamwidth,
            gate_spacing=rng_res,
        )
    else:
        warn(
            "Radar volume of type "
            + radar_out.scan_type
            + ". Only volumes of type PPI or RHI are allowed"
        )
        return None, None

    del radar_out.fields[refl_field]
    del radar_out.fields[width_field]

    radar_out.fields["turbulence"]["data"] = np.ma.masked_values(
        radar_out.fields["turbulence"]["data"], -32768.0
    )

    # prepare for exit
    new_dataset = {"radar_out": radar_out}

    return new_dataset, ind_rad


def process_dealias_fourdd(procstatus, dscfg, radar_list=None):
    """
    Dealiases the Doppler velocity field using the 4DD technique
    from Curtis and Houze, 2001

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            "V" or "Vc"
        filt : int. Dataset keyword
            Flag controlling Bergen and Albers filter, 1 = yes, 0 = no.
        sign : int. Dataset keyword
            Sign convention which the radial velocities in the volume created
            from the sounding data will will. This should match the
            convention used in the radar data. A value of 1 represents when
            positive values velocities are towards the radar, -1 represents
            when negative velocities are towards the radar.


    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "dealV" or "dealVc" (if Vc was provided)
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    vel_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if vel_field not in radar.fields:
        warn("Unable to correct Doppler aliasing. Missing velocity field")
        warn(f"Radar available fields are {radar.fields}")
        return None, None

    corr_vel_field = "dealiased_" + vel_field

    if not dscfg["initialized"]:
        # Use phase unwraping method to obtain first guess

        # get user parameters
        interval_splits = dscfg.get("interval_splits", 3)
        skip_between_rays = dscfg.get("skip_between_rays", 100)
        skip_along_ray = dscfg.get("skip_along_ray", 100)
        centered = dscfg.get("centered", True)

        corr_vel_dict = pyart.correct.dealias_region_based(
            radar,
            ref_vel_field=None,
            interval_splits=interval_splits,
            interval_limits=None,
            skip_between_rays=skip_between_rays,
            skip_along_ray=skip_along_ray,
            centered=centered,
            nyquist_vel=None,
            check_nyquist_uniform=True,
            gatefilter=False,
            rays_wrap_around=None,
            keep_original=False,
            set_limits=False,
            vel_field=vel_field,
            corr_vel_field=corr_vel_field,
        )

        dscfg["initialized"] = 1
    else:
        # get user parameters
        filt = dscfg.get("filt", 1)
        sign = dscfg.get("sign", 1)
        keep_mask = dscfg.get("keep_mask", True)

        last_radar = dscfg["global_data"]

        corr_vel_dict = pyart.correct.dealias_fourdd(
            radar,
            last_radar=last_radar,
            sonde_profile=None,
            gatefilter=False,
            filt=filt,
            rsl_badval=131072.0,
            keep_original=False,
            set_limits=False,
            vel_field=vel_field,
            corr_vel_field=corr_vel_field,
            last_vel_field=corr_vel_field,
            debug=False,
            sign=sign,
        )

        if keep_mask:
            mask = np.ma.getmaskarray(radar.fields[vel_field]["data"])
            corr_vel_dict["data"] = np.ma.masked_where(mask, corr_vel_dict["data"])

    # prepare for exit
    radar_out = deepcopy(radar)
    radar_out.fields = dict()
    radar_out.add_field(corr_vel_field, corr_vel_dict)
    new_dataset = {"radar_out": radar_out}

    # keep current corrected Doppler velocity field in memory
    dscfg["global_data"] = radar_out

    return new_dataset, ind_rad


def process_dealias_region_based(procstatus, dscfg, radar_list=None):
    """
    Dealiases the Doppler velocity field using a region based algorithm

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain,
            "V" or "Vc"
        interval_splits : int, optional
            Number of segments to split the nyquist interval into when finding
            regions of similar velocity.  More splits creates a larger number
            of initial regions which takes longer to process but may result in
            better dealiasing.  The default value of 3 seems to be a good
            compromise between performance and artifact free dealiasing. This
            value is not used if the interval_limits parameter is not None.
        skip_between_rays, skip_along_ray : int, optional
            Maximum number of filtered gates to skip over when joining
            regions, gaps between region larger than this will not be
            connected. Parameters specify the maximum number of filtered gates
            between and along a ray. Set these parameters to 0 to disable
            unfolding across filtered gates.
        centered : bool, optional
            True to apply centering to each sweep after the dealiasing
            algorithm so that the average number of unfolding is near 0. False
            does not apply centering which may results in individual sweeps
            under or over folded by the nyquist interval.
        nyquist_vel : float, optional
            Nyquist velocity of the aquired radar velocity.
            Usually this parameter is provided in the
            Radar object intrument_parameters. If it is not available it can
            be provided as a keyword here.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "dealV" or "dealVc" (if Vc was provided)
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    vel_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if vel_field not in radar.fields:
        warn("Unable to correct Doppler aliasing. Missing velocity field")
        warn(f"Radar available fields are {radar.fields}")
        return None, None

    corr_vel_field = "dealiased_" + vel_field

    # get user parameters
    interval_splits = dscfg.get("interval_splits", 3)
    skip_between_rays = dscfg.get("skip_between_rays", 100)
    skip_along_ray = dscfg.get("skip_along_ray", 100)
    centered = dscfg.get("centered", True)
    nyquist_vel = dscfg.get("nyquist_vel", None)

    corr_vel_dict = pyart.correct.dealias_region_based(
        radar,
        ref_vel_field=None,
        interval_splits=interval_splits,
        interval_limits=None,
        skip_between_rays=skip_between_rays,
        skip_along_ray=skip_along_ray,
        centered=centered,
        nyquist_vel=nyquist_vel,
        check_nyquist_uniform=True,
        gatefilter=False,
        rays_wrap_around=None,
        keep_original=False,
        set_limits=False,
        vel_field=vel_field,
        corr_vel_field=corr_vel_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(corr_vel_field, corr_vel_dict)

    return new_dataset, ind_rad


def process_dealias_unwrap_phase(procstatus, dscfg, radar_list=None):
    """
    Dealiases the Doppler velocity field using multi-dimensional phase
    unwrapping

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain,
            "V" or "Vc"
        unwrap_unit : {'ray', 'sweep', 'volume'}, optional
            Unit to unwrap independently.  'ray' will unwrap each ray
            individually, 'sweep' each sweep, and 'volume' will unwrap the
            entire volume in a single pass.  'sweep', the default, often gives
            superior results when the lower sweeps of the radar volume are
            contaminated by clutter. 'ray' does not use the gatefilter
            parameter and rays where gates ared masked will result in poor
            dealiasing for that ray.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "dealV" or "dealVc" (if Vc was provided)
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    vel_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if vel_field not in radar.fields:
        warn("Unable to correct Doppler aliasing. Missing data")
        return None, None

    corr_vel_field = "dealiased_" + vel_field

    # get user parameters
    unwrap_unit = dscfg.get("unwrap_unit", "sweep")

    corr_vel_dict = pyart.correct.dealias_unwrap_phase(
        radar,
        unwrap_unit=unwrap_unit,
        nyquist_vel=None,
        check_nyquist_uniform=True,
        gatefilter=False,
        rays_wrap_around=None,
        keep_original=False,
        set_limits=False,
        vel_field=vel_field,
        corr_vel_field=corr_vel_field,
        skip_checks=False,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(corr_vel_field, corr_vel_dict)

    return new_dataset, ind_rad


def process_radial_velocity(procstatus, dscfg, radar_list=None):
    """
    Estimates the radial velocity respect to the radar from the wind velocity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            WIND_SPEED, and,
            WIND_DIRECTION, and,
            wind_vel_v
        latitude, longitude : float
            arbitrary coordinates [deg] from where to compute the radial
            velocity. If any of them is None it will be the radar position
        altitude : float
            arbitrary altitude [m MSL] from where to compute the radial
            velocity. If None it will be the radar altitude
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "V"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    v_speed_field = None
    h_speed_field = None
    h_dir_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == "wind_vel_v":
            v_speed_field = get_fieldname_pyart(datatype)
        if datatype == "WIND_SPEED":
            h_speed_field = get_fieldname_pyart(datatype)
        if datatype == "WIND_DIRECTION":
            h_dir_field = get_fieldname_pyart(datatype)

    if h_speed_field is None or h_dir_field is None:
        warn(
            "Horizontal wind speed and direction fields required"
            " to estimate radial velocity"
        )
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if h_speed_field not in radar.fields or h_dir_field not in radar.fields:
        warn("Unable to estimate radial velocity. " "Missing horizontal wind")
        return None, None

    h_speed = radar.fields[h_speed_field]["data"]
    h_dir = radar.fields[h_dir_field]["data"]

    if v_speed_field is None or v_speed_field not in radar.fields:
        warn("Unknown vertical wind speed. Assumed 0")

        if v_speed_field is None:
            v_speed_field == "vertical_wind_component"

        v_speed = np.ma.zeros((radar.nrays, radar.ngates))
    else:
        v_speed = radar.fields[v_speed_field]["data"]

    # user defined parameters
    lat = dscfg.get("latitude", None)
    lon = dscfg.get("longitude", None)
    alt = dscfg.get("altitude", None)

    # get u and v wind components
    h_dir_rad = np.deg2rad(h_dir)
    speed_h_u = h_speed * np.sin(h_dir_rad)  # eastward component
    speed_h_v = h_speed * np.cos(h_dir_rad)  # northward component

    if lat is not None or lon is not None or alt is not None:
        # get antenna coordinates respect to new radar location
        if lat is None:
            lat = radar.latitude["data"][0]
        if lon is None:
            lon = radar.longitude["data"][0]
        if alt is None:
            alt = radar.altitude["data"][0]

        x, y = pyart.core.geographic_to_cartesian_aeqd(
            radar.gate_longitude["data"], radar.gate_latitude["data"], lon, lat
        )
        z = radar.gate_altitude["data"] - alt
        _, azimuths, elevations = pyart.core.cartesian_to_antenna(x, y, z)
        azi_2D_rad = np.deg2rad(azimuths)
        ele_2D_rad = np.deg2rad(elevations)
    else:
        azi_2D_rad = np.broadcast_to(
            np.deg2rad(radar.azimuth["data"])[:, np.newaxis],
            (radar.nrays, radar.ngates),
        )
        ele_2D_rad = np.broadcast_to(
            np.deg2rad(radar.elevation["data"])[:, np.newaxis],
            (radar.nrays, radar.ngates),
        )

    r_speed = pyart.config.get_metadata("velocity")

    # assuming no vertical velocity
    # r_speed['data'] = h_speed*np.cos(h_dir_rad-azi_2D_rad)*np.cos(ele_2D_rad)

    # with vertical velocity included
    r_speed["data"] = (
        speed_h_u * np.sin(azi_2D_rad) + speed_h_v * np.cos(azi_2D_rad)
    ) * np.cos(ele_2D_rad) + np.sin(ele_2D_rad) * v_speed

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field("velocity", r_speed)

    return new_dataset, ind_rad


def process_wind_vel(procstatus, dscfg, radar_list=None):
    """
    Estimates the horizontal or vertical component of the wind from the
    radial velocity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            "V" or "Vc"
        vert_proj : Boolean
            If true the vertical projection is computed. Otherwise the
            horizontal projection is computed
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "wind_vel_h_az", (if vert_proj is False), or,
        "wind_vel_v" (if vert_proj is True)
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    vel_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if vel_field not in radar.fields:
        warn("Unable to retrieve wind speed. Missing data")
        return None, None

    vert_proj = dscfg.get("vert_proj", False)
    wind_field = "azimuthal_horizontal_wind_component"
    if vert_proj:
        wind_field = "vertical_wind_component"

    wind = pyart.retrieve.est_wind_vel(
        radar, vert_proj=vert_proj, vel_field=vel_field, wind_field=wind_field
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(wind_field, wind)

    return new_dataset, ind_rad


def process_windshear(procstatus, dscfg, radar_list=None):
    """
    Estimates the wind shear from the wind velocity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type
        az_tol : float
            The tolerance in azimuth when looking for gates on top
            of the gate when computation is performed

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "windshear_v"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    wind_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if wind_field not in radar.fields:
        warn("Unable to retrieve wind shear. Missing data")
        return None, None

    az_tol = dscfg.get("az_tol", 0.5)
    windshear_field = "vertical_wind_shear"

    windshear = pyart.retrieve.est_vertical_windshear(
        radar, az_tol=az_tol, wind_field=wind_field, windshear_field=windshear_field
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(windshear_field, windshear)

    return new_dataset, ind_rad


def process_windshear_lidar(procstatus, dscfg, radar_list=None):
    """
    Estimates the wind shear from the wind velocity of lidar scans

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            "V" or "Vc"
        az_tol : float
            The tolerance in azimuth when looking for gates on top
            of the gate when computation is performed

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "windshear_v"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    wind_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if wind_field not in radar.fields:
        warn("Unable to retrieve wind shear. Missing data")
        return None, None

    az_tol = dscfg.get("az_tol", 0.5)
    windshear_field = "vertical_wind_shear"

    windshear = pyart.retrieve.est_vertical_windshear_lidar(
        radar, az_tol=az_tol, wind_field=wind_field, windshear_field=windshear_field
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(windshear_field, windshear)

    return new_dataset, ind_rad


def process_vad(procstatus, dscfg, radar_list=None):
    """
    Estimates vertical wind profile using the VAD (velocity Azimuth Display)
    technique

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            "V" or "Vc"
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "wind_vel_h_u", "wind_vel_h_v", "wind_vel_v",
        "estV", "stdV", and "diffV"
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    vel_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if radar_list[ind_rad] is None:
        warn("No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if vel_field not in radar.fields:
        warn("Unable to retrieve wind speed. Missing data")
        return None, None

    # User defined parameters
    npoints_min = dscfg.get("npoints_min", 6)
    azi_spacing_max = dscfg.get("azi_spacing_max", 45.0)
    vel_diff_max = dscfg.get("vel_diff_max", 10.0)

    (u_vel_dict, v_vel_dict, w_vel_dict, vel_est_dict, vel_std_dict, vel_diff_dict) = (
        pyart.retrieve.est_wind_profile(
            radar,
            npoints_min=npoints_min,
            azi_spacing_max=azi_spacing_max,
            vel_diff_max=vel_diff_max,
            rad_vel_field=vel_field,
            u_vel_field="eastward_wind_component",
            v_vel_field="northward_wind_component",
            w_vel_field="vertical_wind_component",
            vel_est_field="retrieved_velocity",
            vel_std_field="retrieved_velocity_std",
            vel_diff_field="velocity_difference",
        )
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(radar)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field("eastward_wind_component", u_vel_dict)
    new_dataset["radar_out"].add_field("northward_wind_component", v_vel_dict)
    new_dataset["radar_out"].add_field("vertical_wind_component", w_vel_dict)
    new_dataset["radar_out"].add_field("retrieved_velocity", vel_est_dict)
    new_dataset["radar_out"].add_field("retrieved_velocity_std", vel_std_dict)
    new_dataset["radar_out"].add_field("velocity_difference", vel_diff_dict)

    return new_dataset, ind_rad


def process_dda(procstatus, dscfg, radar_list=None):
    """
    Estimates horizontal wind speed and direction with a multi-doppler approach
    This method uses the python package pyDDA

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The input data type, must contain
            "V" or "Vc", and,
            "dBuZ", "dBZ", or "dBZc"

        gridconfig : dictionary. Dataset keyword
            Dictionary containing some or all of this keywords:
            xmin, xmax, ymin, ymax, zmin, zmax : floats
                minimum and maximum horizontal distance from grid origin [km]
                and minimum and maximum vertical distance from grid origin [m]
                Defaults -40, 40, -40, 40, 0., 10000.
            latmin, latmax, lonmin, lonmax : floats
                minimum and maximum latitude and longitude [deg], if specified
                xmin, xmax, ymin, ymax will be ignored
            hres, vres : floats
                horizontal and vertical grid resolution [m]
                Defaults 1000., 500.
            latorig, lonorig, altorig : floats
                latitude and longitude of grid origin [deg] and altitude of
                grid origin [m MSL]
                Defaults the latitude, longitude and altitude of the radar
        wfunc : str. Dataset keyword
            the weighting function used to combine the radar gates close to a
            grid point. Possible values BARNES, BARNES2, CRESSMAN, NEAREST
            Default NEAREST
        roif_func : str. Dataset keyword
            the function used to compute the region of interest.
            Possible values: dist_beam, constant
        roi : float. Dataset keyword
            the (minimum) radius of the region of interest in m. Default half
            the largest resolution
        beamwidth : float. Dataset keyword
            the radar antenna beamwidth [deg]. If None that of the key
            radar_beam_width_h in attribute instrument_parameters of the radar
            object will be used. If the key or the attribute are not present
            a default 1 deg value will be used
        beam_spacing : float. Dataset keyword
            the beam spacing, i.e. the ray angle resolution [deg]. If None,
            that of the attribute ray_angle_res of the radar object will be
            used. If the attribute is None a default 1 deg value will be used
        signs : list of integers
            The sign of the velocity field for every radar object.
            A value of 1 represents when
            positive values velocities are towards the radar, -1 represents
            when negative velocities are towards the radar.
        Co : float
            Weight for cost function related to observed radial velocities.
            Default: 1.
        Cm : float
            Weight for cost function related to the mass continuity equation.
            Default: 1500.
        Cx: float
            Smoothing coefficient for x-direction
        Cy: float
            Smoothing coefficient for y-direction
        Cz: float
            Smoothing coefficient for z-direction
        Cb: float
            Coefficient for sounding constraint
        Cv: float
            Weight for cost function related to vertical vorticity equation.
        Cmod: float
            Coefficient for model constraint
        Cpoint: float
            Coefficient for point constraint
        wind_tol: float
            Stop iterations after maximum change in winds is less than this
            value.
        frz : float
            The freezing level in meters. This is to tell PyDDA where to use
            ice particle fall speeds in the wind retrieval verus liquid.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
            "wind_vel_h_u", "wind_vel_h_v" and "wind_vel_v"
    ind_rad : int
        radar index

    """
    if not _PYDDA_AVAILABLE:
        warn("PyDDA and xarray package not available. Unable to compute wind fields")
        return None, None

    if procstatus != 1:
        return None, None

    if len(radar_list) < 2:
        warn("DDA requires data from at least two different radars")
        return None, None

    # check how many radars have to be compared and which datatype to use
    ind_radar_list = []
    field_name_list = []
    datatype_list = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name = get_fieldname_pyart(datatype)
        ind_radar_list.append(int(radarnr[5:8]) - 1)
        datatype_list.append(datatype)
        field_name_list.append(field_name)
    ind_radar_list = np.array(ind_radar_list)
    datatype_list = np.array(datatype_list)
    field_name_list = np.array(field_name_list)

    # Get DDA numerical parameters
    Co = dscfg.get("Co", 1.0)
    Cm = dscfg.get("Cm", 1500.0)
    Cx = dscfg.get("Cx", 0.0)
    Cy = dscfg.get("Cy", 0.0)
    Cz = dscfg.get("Cz", 0.0)
    Cb = dscfg.get("Cb", 0.0)
    Cv = dscfg.get("Cv", 0.0)
    frz = dscfg.get("frz", 4500.0)
    Cmod = dscfg.get("Cmod", 0.0)
    Cpoint = dscfg.get("Cpoint", 0.0)
    wind_tol = dscfg.get("wind_tol", 0.1)

    signs = dscfg.get("signs", [1] * len(set(ind_radar_list)))
    sounding = dscfg.get("sounding", None)
    if sounding:
        dtime = pyart.graph.common.generate_radar_time_begin(radar_list[0])
        sounding_data = read_radiosounding_igra(sounding, dtime)
    else:
        sounding_data = None

    if sounding_data is not None:
        # Remove nan from souding
        sounding_data = sounding_data.dropna()
        # create wind profile from sounding
        wind_prof = pyart.core.HorizontalWindProfile(
            sounding_data["GPH"], sounding_data["WSPD"], sounding_data["WDIR"]
        )
    else:
        # Compute VAD for every radar to initialize DDA
        # z-vector for vad
        z_want = np.arange(
            dscfg["gridConfig"]["zmin"],
            dscfg["gridConfig"]["zmax"] + dscfg["gridConfig"]["vres"],
            dscfg["gridConfig"]["vres"],
        )

        wind_prof = compute_average_vad(
            radar_list,
            z_want,
            signs,
            dscfg["gridConfig"]["lonorig"],
            dscfg["gridConfig"]["latorig"],
        )

    # Get name of reflectivity and velocity fields for each radar
    refl_fields = []
    vel_fields = []
    for i, radar in enumerate(radar_list):
        # Find vel and refl fields
        field_names_rad = field_name_list[ind_radar_list == i]
        vel_field = field_names_rad[
            ["velocity" in vname for vname in field_name_list[ind_radar_list == i]]
        ][0]
        vel_fields.append(vel_field)
        refl_field = field_names_rad[
            ["reflectivity" in vname for vname in field_name_list[ind_radar_list == i]]
        ][0]
        refl_fields.append(refl_field)

    # Grid the variables
    grids = []
    for i, radar in enumerate(radar_list):
        # Now we grid
        dscfg_grid = deepcopy(dscfg)
        dscfg_grid["datatype"] = np.array(dscfg["datatype"])[ind_radar_list == i]
        grids.append(process_grid(1, dscfg_grid, radar_list)[0]["radar_out"])
    grids_pyDDA = []

    # Harmonize variables
    for i, grid in enumerate(grids):
        grid.fields["velocity"] = grid.fields.pop(vel_fields[i])
        if signs[i] == 1:
            grid.fields["velocity"]["data"] *= -1
        grid.fields["reflectivity"] = grid.fields.pop(refl_fields[i])
        grids_pyDDA.append(pydda.io.read_from_pyart_grid(grid))

    # DDA initialization
    grids_pyDDA[0] = pydda.initialization.make_wind_field_from_profile(
        grids_pyDDA[0], wind_prof, vel_field="velocity"
    )

    # Actual DDA computation
    new_grids, params = pydda.retrieval.get_dd_wind_field(
        grids_pyDDA,
        vel_name="velocity",
        refl_field="reflectivity",
        mask_outside_opt=True,
        engine="scipy",
        Co=Co,
        Cm=Cm,
        Cx=Cx,
        Cy=Cy,
        Cz=Cz,
        Cb=Cb,
        Cv=Cv,
        Cmod=Cmod,
        Cpoint=Cpoint,
        wind_tol=wind_tol,
        frz=frz,
    )

    for i, grid in enumerate(new_grids):
        # convert back to pyart grids
        new_grids[i] = convert_pydda_to_pyart_grid(grid)
        # add radar name
        new_grids[i].radar_name = grids[i].radar_name

    # pyDDA returns as many grid objects as input radars
    # the idea here is to merge these grid objects into one
    # and to replace the homogeneized vel and refl fields by the
    # original ones

    # create merged grid from first dda grid
    # First radar will have normal field names
    # Additional radars will have _radar1, radar_2, ..., radar_N
    # appended to their field names
    merged_grid = deepcopy(new_grids[0])
    for var in list(merged_grid.fields):
        if var in grids[0].fields:
            if var == "velocity":
                new_key = vel_fields[0]
            elif var == "reflectivity":
                new_key = refl_fields[0]
            else:
                new_key = var
            merged_grid.fields[new_key] = merged_grid.fields.pop(var)

    # add the other dda grids
    radar_metadata = []
    for i, additional_grid in enumerate(new_grids[1:]):
        for var in list(additional_grid.fields):
            if var in grids[i + 1].fields:
                if var == "velocity":
                    new_key = vel_fields[i + 1]
                elif var == "reflectivity":
                    new_key = refl_fields[i + 1]
                else:
                    new_key = var
                new_key += "_radar{:d}".format(i + 1)
                merged_grid.fields[new_key] = additional_grid.fields[var]

        # Get additional radar metadata
        mdata = {}
        mdata["radar_longitude"] = additional_grid.radar_longitude["data"]
        mdata["radar_latitude"] = additional_grid.radar_latitude["data"]
        # Try to find name of radar
        if additional_grid.radar_name["data"] != "":
            mdata["radar_name"] = additional_grid.radar_name["data"][0]
        elif dscfg["RadarName"][i + 1] != "":
            mdata["radar_name"] = dscfg["RadarName"][i + 1]
        else:
            mdata["radar_name"] = "Unknown"

        radar_metadata.append(mdata)

    # Rename DDA u, v, w fields to pyart names
    merged_grid.fields["eastward_wind_component"] = merged_grid.fields.pop("u")
    merged_grid.fields["northward_wind_component"] = merged_grid.fields.pop("v")
    merged_grid.fields["vertical_wind_component"] = merged_grid.fields.pop("w")

    # Add coordinates of additional radars in metadata
    merged_grid.metadata["additional_radars"] = radar_metadata

    new_dataset = {}
    new_dataset["radar_out"] = merged_grid
    return new_dataset, ind_radar_list




# THIS FILE IS GENERATED FROM PYRAD_PROC SETUP.PY
short_version = '2.0.3'
version = '2.0.3'
full_version = '2.0.3'
git_revision = '7cb27317b16fa51ad6269d70739fd33d941fd413'
compile_date_time = '2025-05-20 07:35'
username = 'runner'
release = True

if not release:
    version = full_version



"""
Pyrad: The Python Radar Toolkit
=====================================

"""

# Detect if we're being called as part of Pyrad's setup procedure
try:
    __PYRAD_SETUP__
except NameError:
    __PYRAD_SETUP__ = False

if __PYRAD_SETUP__:
    import sys as _sys

    _sys.stderr.write("Running from Pyrad source directory.\n")
    del _sys
else:

    # Make sure that deprecation warnings get printed by default
    import warnings as _warnings

    _warnings.simplefilter("always", DeprecationWarning)

    # print out helpful message if build fails or importing from source tree
    # fvj built not checked for the moment
    # from . import __check_build

    # import subpackages
    from . import graph  # noqa
    from . import io  # noqa
    from . import proc  # noqa
    from . import prod  # noqa
    from . import util  # noqa
    from . import flow  # noqa

    # root level functions
    # non at the moment
